"""Definition of the ModelConfiguration class."""

import os
from typing import Callable, Dict, List, Optional

import toml
from pydantic import BaseModel, Field, root_validator, validator
from src import stan_input_functions

AVAILABLE_MODES = ["prior", "posterior", "kfold"]
HERE = os.path.dirname(os.path.abspath(__file__))
STAN_DIR = os.path.join(HERE, "stan")
DEFAULT_DIMS = {"llik": ["observation"], "yrep": ["observation"]}
DEFAULT_SAMPLE_KWARGS = {"show_progress": False}


class InferenceConfiguration(BaseModel):
    """Container for a path to a Stan model and some configuration.

    For example, you may want to compare how well two stan programs fit the
    same data, or how well the same model fits the data with different
    covariates.

    :param name: string name identifying the model configuration

    :param stan_file: Path to a Stan program, with "/" even on windows

    :param prepared_data_dir: Path to a directory containing prepared data

    :param stan_input_function: function from src.stan_input_functions used to
    get a Stan input dictionary from a PreparedData object.

    :param sample_kwargs: dictionary of keyword arguments to
    cmdstanpy.CmdStanModel.sample.

    :param modes: which modes to run the model in. Choose one or more of the
    AVAILABLE_MODES.

    :param dims: map from parameter names to lists of coordinate names. See the
    field names of the file "coords.json" in the prepared_data_dir for possible
    coordinate names.

    :param kfold_folds: How many kfold folds to run

    :param cpp_options: valid choices for the `cpp_options` argument to
    CmdStanModel

    :param stanc_options: valid choices for the `cpp_options` argument to
    CmdStanModel

    """

    name: str
    stan_file: str
    prepared_data_dir: str
    modes: List[str]
    stan_input_function: Callable
    sample_kwargs: dict = Field(default_factory=lambda: DEFAULT_SAMPLE_KWARGS)
    dims: Dict[str, List[str]] = Field(default_factory=lambda: DEFAULT_DIMS)
    kfold_folds: Optional[int] = None
    cpp_options: Optional[dict] = None
    stanc_options: Optional[dict] = None

    def __init__(self, **data):
        """Initialise an InferenceConfiguration."""
        data["stan_input_function"] = getattr(
            stan_input_functions, data["stan_input_function"]
        )
        super().__init__(**data)

    @root_validator
    def check_folds(cls, values):
        """Check that there is a number of folds if required."""
        if "kfold" in values["modes"] and values["kfold_folds"] is None:
            raise ValueError("Set kfold_folds in order to run in kfold mode.")
        return values

    @validator("stan_file")
    def check_stan_file_exists(cls, v):
        """Check that the stan file exists."""
        if not os.path.exists(os.path.join("src", "stan", v)):
            raise ValueError(f"{v} is not a file in src/stan.")
        return v

    @validator("modes")
    def check_modes(cls, v):
        """Check that the provided modes exist."""
        for mode in v:
            if mode not in AVAILABLE_MODES:
                raise ValueError(
                    f"{mode} not in available modes: {AVAILABLE_MODES}."
                )
        return v


def load_inference_configuration(path: str):
    """Load an inference configuration object from a toml file."""
    kwargs = toml.load(path)
    for k, default in zip(
        ["dims", "sample_kwargs"], [DEFAULT_DIMS, DEFAULT_SAMPLE_KWARGS]
    ):
        if k in kwargs.keys():
            kwargs[k] = default | kwargs[k]
    return InferenceConfiguration(**kwargs)



"""Integration tests for functions in src/data_preparation.py."""

from typing import Callable

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from src.data_preparation_functions import (
    prepare_data_interaction,
    prepare_data_no_interaction,
)
from src.util import CoordDict

EXAMPLE_RAW_MEASUREMENTS = pd.DataFrame(
    {
        "X1": [1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2],
        "X2": [1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2],
        "yButIThoughtIdAddSomeLetters": [1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2],
    }
)


@pytest.mark.parametrize(
    "prepare_data_function,name,raw_measurements,expected_measurements,"
    "expected_coords",
    [
        (
            prepare_data_interaction,
            "interaction",
            EXAMPLE_RAW_MEASUREMENTS,
            pd.DataFrame(
                {
                    "x1": [1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2],
                    "x2": [1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2],
                    "y": [1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2],
                    "x1:x2": [1, 4, 9, 1, 4, 9, 1, 4, 9, 1, 4],
                }
            ).apply(lambda s: s.astype(float)),
            {
                "covariate": ["x1", "x2", "x1:x2"],
                "observation": [str(i) for i in range(11)],
            },
        ),
        (
            prepare_data_no_interaction,
            "no_interaction",
            EXAMPLE_RAW_MEASUREMENTS,
            pd.DataFrame(
                {
                    "x1": [1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2],
                    "x2": [1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2],
                    "y": [1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2],
                    "x1:x2": [1, 4, 9, 1, 4, 9, 1, 4, 9, 1, 4],
                }
            ).apply(lambda s: s.astype(float)),
            {
                "covariate": ["x1", "x2"],
                "observation": [str(i) for i in range(11)],
            },
        ),
    ],
)
def test_prepare_data_function(
    prepare_data_function: Callable,
    name: str,
    raw_measurements: pd.DataFrame,
    expected_measurements: pd.DataFrame,
    expected_coords: CoordDict,
):
    """Check that a prepare data function behaves as expected."""
    prepped = prepare_data_function(raw_measurements)
    assert prepped.name == name
    assert prepped.coords == expected_coords
    assert_frame_equal(prepped.measurements, expected_measurements)



"""Dataclasses and functions for running bibat's command line wizard."""

from typing import Callable, Dict, List, Optional, Union

import click
from pydantic import root_validator
from pydantic.dataclasses import dataclass


@dataclass
class WizardStr:
    """A string field."""

    name: str
    prompt: str
    default: Optional[str] = None
    default_function: Optional[Callable] = None

    @root_validator
    def default_or_default_function_exists(cls, values):
        """Either the default or the default function must not be None."""
        assert (values["default"] is not None) or (
            values["default_function"] is not None
        ), "Either the default or the default function must not be None."
        return values


@dataclass
class WizardChoice:
    """A choice field."""

    name: str
    prompt: str
    options: List[str]
    default: Optional[str] = None
    default_function: Optional[Callable] = None

    @root_validator
    def default_is_an_option(cls, values):
        """Check that the default is one of the options."""
        if isinstance(values["default"], str):
            msg = (
                f"default {values['default']} "
                f"not in options {values['options']}"
            )
            assert values["default"] in values["options"], msg
            return values

    @root_validator
    def default_or_default_function_exists(cls, values):
        """Either the default or the default function must not be None."""
        assert (values["default"] is not None) or (
            values["default_function"] is not None
        ), "Either the default or the default function must not be None."
        return values


def prompt_user(
    wf: Union[WizardStr, WizardChoice], context: Optional[Dict]
) -> str:
    """Prompt the user for an input and parse it with click."""
    if context is not None and wf.default_function is not None:
        default = wf.default_function(context)
    elif isinstance(wf.default, str):
        default = wf.default
    else:
        raise ValueError(f"wf.default has unexpected type {type(wf.default)}")
    if isinstance(wf, WizardStr):
        return click.prompt(wf.prompt, default=default, type=str)
    elif isinstance(wf, WizardChoice):
        return click.prompt(
            wf.prompt,
            default=default,
            type=click.Choice(wf.options),
            show_choices=True,
        )
    else:
        raise ValueError(f"input {wf} is not a WizardStr or a WizardChoice")



"""Read the data in RAW_DIR and save prepared data to PREPARED_DIR."""

import os

import pandas as pd
from src import data_preparation_functions
from src.prepared_data import write_prepared_data

DATA_PREPARATION_FUNCTIONS_TO_RUN = [
    data_preparation_functions.prepare_data_fake_interaction,
    data_preparation_functions.prepare_data_interaction,
    data_preparation_functions.prepare_data_no_interaction,
]

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
RAW_DIR = os.path.join(DATA_DIR, "raw")
PREPARED_DIR = os.path.join(DATA_DIR, "prepared")
RAW_DATA_FILES = {
    "raw_measurements": os.path.join(RAW_DIR, "raw_measurements.csv"),
}


def main():
    """Save prepared data in the PREPARED_DIR."""
    print("Reading raw data...")
    raw_data = {
        k: pd.read_csv(v, index_col=None) for k, v in RAW_DATA_FILES.items()
    }
    print("Preparing data...")
    for dpf in DATA_PREPARATION_FUNCTIONS_TO_RUN:
        print(f"Running data preparation function {dpf.__name__}...")
        prepared_data = dpf(raw_data["raw_measurements"])
        output_dir = os.path.join(PREPARED_DIR, prepared_data.name)
        print(f"\twriting files to {output_dir}")
        if not os.path.exists(PREPARED_DIR):
            os.mkdir(PREPARED_DIR)
        write_prepared_data(prepared_data, output_dir)


if __name__ == "__main__":
    main()



"""Run all the configs in the model_configurations folder."""

import os
from typing import Dict

import arviz as az
import cmdstanpy
import numpy as np
import xarray as xr
from sklearn.model_selection import KFold
from src.inference_configuration import (
    AVAILABLE_MODES,
    InferenceConfiguration,
    load_inference_configuration,
)
from src.prepared_data import load_prepared_data
from src.util import CoordDict

RUNS_DIR = "inferences"
STAN_DIR = os.path.join("src", "stan")


def get_sample_kwargs(raw: Dict, mode: str) -> Dict:
    """Get the sampel kwargs for a mode from raw toml."""
    out = {k: v for k, v in raw.items() if k not in AVAILABLE_MODES}
    if mode in raw.keys():
        out |= raw[mode]
    return out


def get_kfold_llik(
    ic: InferenceConfiguration,
    stan_input_base: Dict,
    model: cmdstanpy.CmdStanModel,
    coords: CoordDict,
) -> xr.DataArray:
    """Run k-fold cross-validation and return out-of-sample log likelihood.

    :param ic: InferenceConfiguration, must have non-null field kfold_folds

    :param stan_input_base: stan input dictionary

    :param model: CmdStanModel
    """
    n_folds = ic.kfold_folds
    sample_kwargs = get_sample_kwargs(ic.sample_kwargs, "kfold")
    kf = KFold(n_folds, shuffle=True, random_state=1234)
    lliks_by_fold = []
    full_ix = np.array(stan_input_base["ix_train"])
    for fold, (ix_train, ix_test) in enumerate(kf.split(full_ix)):
        input_dict = stan_input_base | {
            "likelihood": 1,
            "N_train": len(ix_train),
            "N_test": len(ix_test),
            "ix_train": full_ix[ix_train].tolist(),
            "ix_test": full_ix[ix_test].tolist(),
        }
        mcmc = model.sample(data=input_dict, **sample_kwargs)
        llik_fold = mcmc.draws_xr(vars=["llik"])
        # remember the fold
        llik_fold["fold"] = fold
        llik_fold = llik_fold.set_coords("fold")
        # set value of index "chain" to zero to match arviz convention
        llik_fold = llik_fold.assign_coords(
            {"new_chain": ("chain", [0])}
        ).set_index(chain="new_chain")
        # get observation correct
        llik_fold["llik_dim_0"] = np.array(coords["observation"])[ix_test]
        # rename coord "llik_dim_0" to "observation"
        llik_fold = llik_fold.rename({"llik_dim_0": "observation"})
        lliks_by_fold.append(llik_fold["llik"])
    return xr.concat(lliks_by_fold, dim="observation").sortby("observation")


def main():
    """Run the main script."""
    run_dirs = [
        os.path.join(RUNS_DIR, d)
        for d in os.listdir(RUNS_DIR)
        if os.path.isdir(os.path.join(RUNS_DIR, d))
    ]
    for run_dir in sorted(run_dirs):
        ic = load_inference_configuration(os.path.join(run_dir, "config.toml"))
        stan_file = os.path.join(STAN_DIR, ic.stan_file)
        prepared_data_dir = os.path.join(
            "data", "prepared", ic.prepared_data_dir
        )
        model = cmdstanpy.CmdStanModel(
            stan_file=stan_file,
            cpp_options=ic.cpp_options,
            stanc_options=ic.stanc_options,
        )
        prepared_data = load_prepared_data(prepared_data_dir)
        stan_input_base = ic.stan_input_function(prepared_data)
        idata_kwargs = {
            "observed_data": stan_input_base,
            "log_likelihood": "llik",
            "coords": prepared_data.coords,
            "dims": ic.dims,
        }
        for mode, likelihood in [("prior", 0), ("posterior", 1)]:
            if mode in ic.modes:
                sample_kwargs = get_sample_kwargs(ic.sample_kwargs, mode)
                input_dict = stan_input_base | {"likelihood": likelihood}
                idata_kwargs[mode] = model.sample(
                    data=input_dict, **sample_kwargs
                )
                idata_kwargs[f"{mode}_predictive"] = "yrep"
        idata = az.from_cmdstanpy(**idata_kwargs)
        if "kfold" in ic.modes:
            llik_kfold = get_kfold_llik(
                ic, stan_input_base, model, prepared_data.coords
            )
            idata.log_likelihood["llik_kfold"] = llik_kfold
        idata_file = os.path.join(run_dir, "idata.json")
        print(f"Saving idata to {idata_file}")
        idata.to_json(idata_file)


if __name__ == "__main__":
    main()



"""Some handy python functions."""

from typing import Dict, List, NewType, Optional, Union

import numpy as np
import pandas as pd

CoordDict = NewType("CoordDict", Dict[str, List[str]])
StanInputDict = Dict["str", Union[int, float, List]]


def one_encode(s: pd.Series) -> pd.Series:
    """Replace a series's values with 1-indexed integer factors.

    :param s: a pandas Series that you want to factorise.

    """
    return pd.Series(pd.factorize(s)[0] + 1, index=s.index)


def make_columns_lower_case(df: pd.DataFrame) -> pd.DataFrame:
    """Make a DataFrame's columns lower case.

    :param df: a pandas DataFrame
    """
    new = df.copy()
    if isinstance(new.columns, pd.MultiIndex):
        new.columns = pd.MultiIndex.from_arrays(
            [
                [c.lower() for c in new.columns.get_level_values(i)]
                for i in range(len(new.columns.levels))
            ]
        )
    else:
        new.columns = pd.Index([c.lower() for c in new.columns])
    return new


def check_is_df(maybe_df) -> pd.DataFrame:
    """Shut up the type checker."""
    assert isinstance(maybe_df, pd.DataFrame)
    return maybe_df


def stanify_dict(d: Dict) -> StanInputDict:
    """Make sure a dictionary is a valid Stan input.

    :param d: input dictionary, possibly with wrong types
    """
    out: StanInputDict = {}
    for k, v in d.items():
        if not isinstance(k, str):
            raise ValueError(f"key {str(k)} is not a string!")
        elif isinstance(v, pd.Series):
            out[k] = v.to_list()
        elif isinstance(v, pd.DataFrame):
            out[k] = v.values.tolist()
        elif isinstance(v, np.ndarray):
            out[k] = v.tolist()
        else:
            out[k] = v
    return out


def standardise(
    s: pd.Series, mu: Optional[float] = None, std: Optional[float] = None
) -> pd.Series:
    """Standardise a series by subtracting mu and dividing by sd."""
    if mu is None:
        mu = s.mean()
    if std is None:
        std = s.std()
    return s.subtract(mu).divide(std)


def center(
    s: pd.Series,
    mu: Optional[float] = None,
) -> pd.Series:
    """Center a series by subtracting mu."""
    if mu is None:
        mu = s.mean()
    return s.subtract(mu)



"""Provides functions prepare_data_x.

These functions should take in a dataframe of measurements and return a
PreparedData object.

"""

import pandas as pd
from src.prepared_data import PreparedData


def prepare_data_2006(measurements_raw: pd.DataFrame) -> PreparedData:
    """Prepare the 2006 data."""
    measurements = measurements_raw.rename(
        columns={"K": "n_attempt", "y": "n_success"}
    ).assign(
        season="2006",
        player=lambda df: [f"2006-player-{i+1}" for i in range(len(df))],
    )
    return PreparedData(
        name="2006",
        coords={
            "player": measurements["player"].tolist(),
            "season": measurements["season"].tolist(),
        },
        measurements=measurements,
    )


def prepare_data_recent(measurements_raw: pd.DataFrame) -> PreparedData:
    """Prepare the recent data.

    There are a few substantive data choices here.

    First, the function excludes players who have a '1' in their position as
    these are likely pitchers, as well as players with fewer than 20 at bats.

    Second, the function defines a successes and attempts according to the
    'on-base percentage' metric, so a success is a time when a player got a hit,
    a base on ball/walk or a hit-by-pitch and an attempt is an at-bat or a
    base-on-ball/walk or a hit-by-pitch or a sacrifice fly. This could have
    alternatively been calculated as just hits divided by at-bats, but my
    understanding is that this method underrates players who are good at getting
    walks.

    """

    def filter_batters(df: pd.DataFrame):
        return (df["AB"] >= 20) & ~df["Pos.Summary"].astype(str).str.contains(
            "1"
        )

    measurements = (
        measurements_raw.assign(
            season=lambda df: df["year"].astype(str),
            player=lambda df: df["Name.additional"]
            .str.cat(df["Tm"])
            .str.cat(df["Lg"])
            .str.cat(df["season"]),
            n_attempt=lambda df: df[["AB", "BB", "HBP", "SF"]]
            .fillna(0)
            .sum(axis=1)
            .astype(int),
            n_success=lambda df: (
                df[["H", "BB", "HBP"]].fillna(0).sum(axis=1).astype(int)
            ),
        )
        .loc[filter_batters, ["player", "season", "n_attempt", "n_success"]]
        .copy()
    )
    return PreparedData(
        name="recent",
        coords={
            "player": measurements["player"].tolist(),
            "season": measurements["season"].tolist(),
        },
        measurements=measurements,
    )



"""Provides functions prepare_data_x.

These functions should take in a dataframe of measurements and return a
PreparedData object.

"""

import numpy as np
import pandas as pd
from src.prepared_data import PreparedData
from src.util import CoordDict, make_columns_lower_case

NEW_COLNAMES = {"yButIThoughtIdAddSomeLetters": "y"}
DROPNA_COLS = ["y"]
N_CV_FOLDS = 10
DIMS = {
    "b": ["covariate"],
    "y": ["observation"],
    "yrep": ["observation"],
    "llik": ["observation"],
}


def prepare_data_interaction(measurements_raw: pd.DataFrame) -> PreparedData:
    """Prepare data with an interaction column."""
    measurements = process_measurements(measurements_raw)
    return PreparedData(
        name="interaction",
        coords=CoordDict(
            {
                "covariate": ["x1", "x2", "x1:x2"],
                "observation": measurements.index.tolist(),
            }
        ),
        measurements=measurements,
    )


def prepare_data_no_interaction(measurements_raw: pd.DataFrame) -> PreparedData:
    """Prepare data with no interaction column."""
    measurements = process_measurements(measurements_raw)
    return PreparedData(
        name="no_interaction",
        coords=CoordDict(
            {
                "covariate": ["x1", "x2"],
                "observation": measurements.index.tolist(),
            }
        ),
        measurements=measurements,
    )


def prepare_data_fake_interaction(
    measurements_raw: pd.DataFrame,
) -> PreparedData:
    """Prepare fake data with an interaction column."""
    TRUE_PARAMS = {"a": 1, "b": [0.6, -0.3, 0.2], "sigma": 0.3}
    x_cols = ["x1", "x2", "x1:x2"]
    measurements = process_measurements(measurements_raw)
    yhat = TRUE_PARAMS["a"] + measurements[x_cols] @ np.array(TRUE_PARAMS["b"])
    measurements["y"] = np.random.normal(
        yhat, TRUE_PARAMS["sigma"]
    )  # type: ignore
    return PreparedData(
        name="fake_interaction",
        coords=CoordDict(
            {
                "covariate": x_cols,
                "observation": measurements.index.tolist(),
            }
        ),
        measurements=measurements,
    )


def process_measurements(measurements: pd.DataFrame) -> pd.DataFrame:
    """Process the measurements.

    This is to illustrate how you might want to do common table manipulation
    tasks like filtering, changing column names and adding new columns.

    Note that if you want, you can use different measurement processing
    functions for different prepare_data functions

    """
    out = (
        measurements.rename(columns=NEW_COLNAMES)
        .pipe(make_columns_lower_case)
        .dropna(subset=DROPNA_COLS, axis=0)
    ).copy()
    for col in ["x1", "x2", "y"]:
        out[col] = out[col].astype(float)
    out["x1:x2"] = out["x1"] * out["x2"]
    return out



"""Cli for constructing the template.

I used this approach because I wanted to make longer prompts.
"""

import os

import click
from cookiecutter.main import cookiecutter

from bibat import __version__ as bibat_version
from bibat.wizarding import WizardChoice, WizardStr, prompt_user

THIS_DIR = os.path.dirname(__file__)
TOOLING_PACKAGES = [
    "pandas-stubs",
    "types-toml",
    "flake8",
    "flake8-bugbear",
    "flake8-docstrings",
    "mypy",
    "python-lsp-server[all]",
    "python-lsp-black",
    "pylsp-mypy",
    "pyls-isort",
]

WIZARD_FIELDS = [
    WizardStr("project_name", "What is your project called?", "Project Name"),
    WizardStr(
        "repo_name",
        "What should the project repository be called",
        default_function=lambda context: context["project_name"]
        .lower()
        .replace(" ", "_"),
    ),
    WizardStr("author_name", "What is your name?", default="Author name"),
    WizardStr("author_email", "What is your email?", default="Author email"),
    WizardStr(
        "coc_contact",
        "Who should be the code of conduct contact?",
        default_function=lambda context: context["author_email"],
    ),
    WizardStr(
        "description",
        "Please briefly describe your project",
        default="A short description of the project.",
    ),
    WizardChoice(
        "open_source_license",
        "Choose an open source license from these options:",
        ["MIT", "BSD-3-Clause", "No license file"],
        default="MIT",
    ),
    WizardChoice(
        "docs_format",
        "How would you like to document your project?",
        ["Quarto", "Sphinx", "No docs"],
        default="Quarto",
    ),
    WizardStr(
        "create_tests_directory",
        "Would you like to create a tests directory?",
        default="y",
    ),
    WizardStr(
        "create_dotgithub_directory",
        "Would you like to create a .github directory?",
        default="y",
    ),
    WizardStr(
        "install_python_tooling",
        "Would you like to install these handy Python tools?\n\t"
        + "\n\t".join(TOOLING_PACKAGES),
        default="y",
    ),
]


@click.command()
@click.option(
    "--config-file",
    type=click.Path(exists=True),
    default=None,
    help="Path to a yaml file with prefilled cookiecutter fields.",
)
def main(config_file):
    """Generate a Bayesian statistical analysis project."""
    if config_file is not None:
        cookiecutter(THIS_DIR, no_input=True, config_file=config_file)
        return
    click.echo("Welcome to the Batteries-Included Bayesian Analysis Template!")
    context = {}
    for wizard_field in WIZARD_FIELDS:
        context[wizard_field.name] = prompt_user(wizard_field, context)
    context["bibat_version"] = bibat_version
    cookiecutter(
        THIS_DIR, no_input=True, extra_context=context, config_file=config_file
    )


if __name__ == "__main__":
    main()



"""Class defining what prepared data looks like."""

import json
import os

import pandas as pd
import pandera as pa
from pydantic.dataclasses import dataclass
from src.util import CoordDict

NAME_FILE = "name.txt"
COORDS_FILE = "coords.json"
MEASUREMENTS_FILE = "measurements.csv"


class MeasurementsDF(pa.SchemaModel):
    """A PreparedData should have a measurements dataframe like this.

    Other columns are also allowed!
    """

    x1: pa.typing.Series[float]
    x2: pa.typing.Series[float]
    x1colonx2: pa.typing.Series[float] = pa.Field(alias="x1:x2")
    y: pa.typing.Series[float]


@dataclass
class PreparedData:
    """What prepared data looks like in this analysis."""

    name: str
    coords: CoordDict
    measurements: pa.typing.DataFrame[MeasurementsDF]


def load_prepared_data(directory: str) -> PreparedData:
    """Load prepared data from files in directory."""
    with open(os.path.join(directory, COORDS_FILE), "r") as f:
        coords = json.load(f)
    with open(os.path.join(directory, NAME_FILE), "r") as f:
        name = f.read()
    measurements = pd.read_csv(os.path.join(directory, MEASUREMENTS_FILE))
    return PreparedData(
        name=name,
        coords=coords,
        measurements=measurements,
    )


def write_prepared_data(prepped: PreparedData, directory):
    """Write prepared data files to a directory."""
    if not os.path.exists(directory):
        os.mkdir(directory)
        prepped.measurements.to_csv(os.path.join(directory, MEASUREMENTS_FILE))
    with open(os.path.join(directory, COORDS_FILE), "w") as f:
        json.dump(prepped.coords, f)
    with open(os.path.join(directory, NAME_FILE), "w") as f:
        f.write(prepped.name)



"""Functions for generating input to Stan from prepared data."""


from typing import Dict

import numpy as np
from scipy.special import expit, logit
from src.prepared_data import PreparedData


def get_stan_input_normal(ppd: PreparedData) -> Dict:
    """General function for creating a Stan input."""
    return {
        "N": len(ppd.measurements),
        "K": ppd.measurements["n_attempt"].values,
        "y": ppd.measurements["n_success"].values,
        "prior_mu": [logit(0.25), 0.2],
        "prior_tau": [0.2, 0.1],
        "prior_b_K": [0, 0.03],
    }


def get_stan_input_gpareto(ppd: PreparedData) -> Dict:
    """General function for creating a Stan input."""
    return {
        "N": len(ppd.measurements),
        "K": ppd.measurements["n_attempt"].values,
        "y": ppd.measurements["n_success"].values,
        "min_alpha": logit(0.1),
        "max_alpha": logit(0.5),
        "prior_sigma": [1.5, 0.4],
        "prior_k": [-0.5, 1],
    }


def get_stan_input_normal_fake(ppd: PreparedData) -> Dict:
    """Generate fake Stan input consistent with the normal model."""
    N = len(ppd.measurements)
    rng = np.random.default_rng(seed=1234)
    true_param_values = {
        "mu": logit(0.25),
        "tau": 0.18,  # 2sds is 0.19 to 0.32 batting average
        "b_K": 0.04,  # slight effect of more attempts
        "alpha_std": rng.random.normal(loc=0, scale=1, size=N),
    }
    K = ppd.measurements["n_attempt"].values
    log_K_std = (np.log(K) - np.log(K).mean()) / np.log(K).std()
    alpha = (
        true_param_values["mu"]
        + true_param_values["b_K"] * log_K_std
        + true_param_values["tau"] * true_param_values["alpha_std"]
    )
    y = rng.random.binomial(K, expit(alpha))
    return {"N": N, "K": K, "y": y} | true_param_values


def get_stan_input_gpareto_fake(ppd: PreparedData) -> Dict:
    """Generate fake Stan input consistent with the gpareto model."""
    N = len(ppd.measurements)
    K = ppd.measurements["n_attempt"].values
    min_alpha = 0.1
    rng = np.random.default_rng(seed=1234)
    true_param_values = {"sigma": -1.098, "k": 0.18}
    true_param_values["alpha"] = gpareto_rvs(
        rng,
        N,
        min_alpha,
        true_param_values["k"],
        true_param_values["sigma"],
    )
    y = rng.binomial(K, expit(true_param_values["alpha"]))
    return {"N": N, "K": K, "y": y, "min_alpha": min_alpha} | true_param_values


def gpareto_rvs(
    rng: np.random.Generator, size: int, mu: float, k: float, sigma: float
):
    """Generate random numbers from a generalised pareto distribution.

    See https://en.wikipedia.org/wiki/Generalized_Pareto_distribution for
    source.

    """
    U = rng.uniform(size)
    if k == 0:
        return mu - sigma * np.log(U)
    else:
        return mu + (sigma * (U**-k) - 1) / sigma



"""Fetch raw data from the internet."""

import os

import pandas as pd
import requests

URLS = {
    "2006": "https://raw.githubusercontent.com/stan-dev/"
    "example-models/master/knitr/pool-binary-trials/baseball-hits-2006.csv",
    "recent": "https://discourse.mc-stan.org/uploads/short-url/"
    "cRJtia2jKxiX01ZEraHUVK7jbHI.RData",
}
OUT_FILES = {
    "2006": os.path.join("data", "raw", "baseball-hits-2006.csv"),
    "recent": os.path.join("data", "raw", "recent.RData"),
}

if __name__ == "__main__":
    print(f"Fetching 2006 data from {URLS['2006']}")
    data_2006 = pd.read_csv(URLS["2006"], comment="#")
    print(f"Writing 2006 data to {OUT_FILES['2006']}")
    data_2006.to_csv(OUT_FILES["2006"])
    print(f"Fetching recent data from {URLS['recent']}")
    with requests.get(URLS["recent"]) as f:
        recent = f.text
    print(f"Writing to {OUT_FILES['recent']}")
    with open(OUT_FILES["recent"], "wb") as f:
        f.write(recent.encode("utf8", errors="replace"))



"""Functions for generating input to Stan from prepared data."""


from typing import Dict, List

import pandas as pd
from src.prepared_data import PreparedData


def get_stan_input(measurements: pd.DataFrame, x_cols: List[str]) -> Dict:
    """General function for creating a Stan input."""
    return {
        "N": len(measurements),
        "N_train": len(measurements),
        "N_test": len(measurements),
        "K": len(x_cols),
        "x": measurements[x_cols].values.tolist(),
        "y": measurements["y"].tolist(),
        "ix_train": [i + 1 for i in range(len(measurements))],
        "ix_test": [i + 1 for i in range(len(measurements))],
    }


def get_stan_input_interaction(prepared_data: PreparedData) -> Dict:
    """Get a Stan input with an interaction predictor."""
    return get_stan_input(prepared_data.measurements, ["x1", "x2", "x1:x2"])


def get_stan_input_no_interaction(prepared_data: PreparedData) -> Dict:
    """Get a Stan input without an interaction predictor."""
    return get_stan_input(prepared_data.measurements, ["x1", "x2"])
